; ************************************************
; * Name:    Star 3D				 *
; * Version: 2.00				 *
; * Purpose: 3D starfield simulation for CueDark *
; * Author:  Marcel Kilgus			 *
; ************************************************

; The code was originally written by me in May 1996 and became known as
; "star_bin". It was a direct translation of one of my Turbo Pascal
; programs, so the code wasn't very clean (and still isn't at some
; locations). The old code used QDOS calls to draw every single star
; which made it quite slow, now it draws directly to the screen.
; The code is optimized for QPC's 68k emulation kernel where possible,
; I didn't really care for real 68ks. The difference shouldn't be great,
; though. Anyway, the underlying concept is a better place to start
; optimizing, it's far from perfect. But I needed to get this code done
; quickly.

; Last change: 23.01.2000

; Star definition block
d1x	equ	0		; word	-1, 0 or 1 x-step if it's a normal step
d1y	equ	2		; word	     "     y              "
d2x	equ	4		; word	-1, 0 or 1 x-step if it's a special step
d2y	equ	6		; word	     "     y              "
x	equ	8		; word	X-position
y	equ	10		; word	Y-position
m	equ	12		; word	No. of steps to the far side of the screen
n	equ	14		; word
s	equ	16		; word
i	equ	18		; word
speed	equ	20		; byte
speedcn equ	21		; byte
actnum	equ	22		; byte
color	equ	23		; byte
active	equ	24		; word
ssize	equ	26		;	size of definition block

	section code
	data	1024		; some space for the stack and stuff

;	 filetype 1

	include 'win1_keys_qdos_io'	; These are Tony Tebby's key-files I
	include 'win1_keys_qdos_ioa'	; took out of the SMSQ/E sources.
	include 'win1_keys_qdos_sms'	; I think you'll have something similar
	include 'win1_keys_sys'
	include 'win1_keys_con'
	include 'win1_mac_config'

	bra	start
	nop
	dc.w	$4afb
	dc.w	24
	dc.b	'Star 3D by Marcel Kilgus'

stars	dc.w	1000
speeds	dc.w	5
scolor	dc.b	0,0

	mkcfhead {Star3D},{2.00}
	mkcfitem word,0,stars,,,{Amount of stars},1,30000
	mkcfitem word,0,speeds,,,{Amount of different speed stages},1,100
	mkcfitem code,0,scolor,,,{Star colour}\
	  0,D,{Depends on speed},1,R,{Red},2,G,{Green},3,W,{White}
	mkcfend

; This code was for testing only
; scr	  dc.w	  15
;	  dc.b	  'scr_800x600a0x0',0

start	move.w	stars(pc),d1
	mulu	#ssize,d1
	moveq	#-1,d2
	moveq	#sms.achp,d0
	trap	#1
	tst.l	d0
	bne.s	endjob
	movea.l a0,a5

	moveq	#sms.info,d0
	trap	#do.smsq
	movea.l a0,a6
	move.l	6(a7),d7	; xsize / ysize
	move.l	2(a7),a0	; channel id

;	   moveq   #$1,d0	 ; For testing we open a little window
;	   moveq   #-1,d1
;	   moveq   #0,d3
;	   lea	   scr,a0
;	   trap    #2
;	   moveq   #iow.spap,d0
;;	    moveq   #3,d1
;	   moveq   #0,d1
;	   moveq   #-1,d3
;	   trap    #3
;	   moveq   #iow.clra,d0  ; Now clear it
;	   moveq   #-1,d3
;	   trap    #3
;	   move.l  #$03200258,d7 ; The size of the window

	move.l	a5,-(a7)	; save base of stars
	move.w	stars(pc),d6	; loop counter for initialisation
	subq.w	#1,d6
init_loop
	bsr	init_star	; init data for this star
	move.w	#1000,d2	; Star is completely ignored while this counter is > 0
	bsr	random		; Without this all stars would start at the same time,
	move.w	d0,active(a5)	; which doesn't really look nice
	add.w	#ssize,a5	; next star
	dbf	d6,init_loop

; ----- Main drawing loop
main_loop
	move.w	stars(pc),d1
	subq.w	#1,d1
	move.l	(a7),a1 	; base of stars
	moveq	#iow.xtop,d0
	lea	draw_all_stars,a2 ; Call routine as a part of the screen driver
	moveq	#-1,d3
	trap	#3
	bra.s	main_loop

draw_all_stars
	movem.l d1-d7/a0-a6,-(a7)
	move.w	d1,d6
	move.l	a1,a5
	move.l	sd_xsize(a0),d7
star_loop
	bsr.s	process_star
	add.w	#ssize,a5	; next star
	dbf	d6,star_loop	; for all stars
	movem.l (a7)+,d1-d7/a0-a6
	moveq	#0,d0
	rts

endjob
	moveq	#sms.frjb,d0
	moveq	#-1,d1
	moveq	#0,d3
	trap	#do.smsq

; ----- Main calculation routine
process_star
	tst.w	active(a5)	; Is the star already active?
	bne.s	not_active_yet

	addq.b	#1,speedcn(a5)	; Increment speed (or wait) counter
	move.b	speedcn(a5),d0
	cmp.b	speed(a5),d0
	ble.s	end_process
	clr.b	speedcn(a5)	; Reset speed/wait counter for next move

	move.l	x(a5),d1	; Old X/Y position
	bsr.s	clear_star

	addq.w	#1,i(a5)	; I += 1

	move.w	m(a5),d1	; M
	cmp.w	i(a5),d1
	ble	init_star	; I > M so the star's out of the screen and
				; it's time to calculate a new one
	move.w	s(a5),d0	; S
	add.w	n(a5),d0	; S += N
	cmp.w	d1,d0
	ble.s	dostep

	sub.w	d1,d0		; S -= M
	move.w	d0,s(a5)
	move.l	d1x(a5),d0
	add.w	d0,y(a5)	; Y += D1Y
	swap	d0
	add.w	d0,x(a5)	; X += D1X
	bra.s	didstep

dostep	move.w	d0,s(a5)
	move.l	d2x(a5),d0
	add.w	d0,y(a5)	; Y += D2Y
	swap	d0
	add.w	d0,x(a5)	; X += D2X
didstep
	move.l	x(a5),d1
	tst.b	actnum(a5)	; to avoid that all stars start in the middle of the screen,
	beq.s	draw_star	;   we let them fly a little bit before we show them
	subq.b	#1,actnum(a5)
end_process
	rts

not_active_yet
	subq.w	#1,active(a5)	; Decrement active-counter
	rts

; ----- Utilities
clear_star
	moveq	#0,d2
	bra.s	paint_star
draw_star
	moveq	#1,d2
paint_star
	tst.w	d1		; Ensure Y is not negative
	bmi.s	draw_end
	tst.l	d1		; Ensure X is not negative
	bmi.s	draw_end
	cmp.w	sd_ysize(a0),d1 ; Ensure Y is within screen limit
	bge.s	draw_end
	movea.l sd_scrb(a0),a1	; Screen base
	move.w	d1,d0
	mulu	sd_linel(a0),d0 ; Line length
	add.l	d0,a1
	swap	d1
	move.w	d1,d0
	andi.w	#$7,d1
	asr.w	#2,d0
	andi.w	#$7FFE,d0	; Always even address
	add.w	d0,a1
	moveq	#7,d3		; Calculate bit of byte for pixel
	sub.w	d1,d3
	moveq	#1,d0
	asl.b	d3,d0
	tst.b	d2
	bne.s	draw_it
	eor.b	#$ff,d0
	and.b	d0,(a1)
	and.b	d0,1(a1)
	rts
draw_it
	move.b	color(a5),d1
	and.b	#2,d1
	beq.s	draw_it2
	or.b	d0,(a1)
draw_it2
	move.b	color(a5),d1
	and.b	#1,d1
	beq.s	draw_end
	or.b	d0,1(a1)
draw_end
	rts

; d7.l : x-size / y-size of screen
; a5.l : pointer to star definition
; a6.l : base of sysvars
init_star
	moveq	#4,d2		; first we have to get a raw direction
	bsr	random

	subq.w	#1,d0		; up?
	bpl.s	dir_1
	move.l	d7,d2
	swap	d2		; xsize
	bsr	random		; 0 <= x < xsize
	move.w	d0,d4
	swap	d4
	clr.w	d4		; y=0
	bra.s	dir_all

dir_1	subq.w	#1,d0		; down?
	bpl.s	dir_2
	move.l	d7,d2
	swap	d2		; xsize
	bsr	random		; 0 <= x < xsize
	move.w	d0,d4
	swap	d4
	move.w	d7,d4		; y=ysize
	bra.s	dir_all

dir_2	subq.w	#1,d0		; left?
	bpl.s	dir_3
	clr.l	d4		; x=0
	move.w	d7,d2
	bsr	random		; 0 <= y < ysize
	move.w	d0,d4
	bra.s	dir_all

dir_3	move.l	d7,d4		; the star flies to the right / x=xsize
	move.w	d7,d2
	bsr	random		; 0 <= y < ysize
	move.w	d0,d4

dir_all move.l	d7,d0		; the screen size
	lsr.w	#1,d0		; y screen center
	move.w	d0,y(a5)	; the beginning point is x/2, y/2
	swap	d0
	lsr.w	#1,d0		; x screen center
	move.w	d0,x(a5)
	swap	d4
	sub.w	d0,d4		; amount of x steps to the ending point of the star
	swap	d0
	swap	d4
	sub.w	d0,d4		; amount of y steps to the ending point of the star

	move.l	d4,d0
	bsr	sgn		; d1y
	swap	d0
	bsr	sgn		; d1x
	swap	d0
	move.l	d0,d1x(a5)	; d1x and d1y
	move.w	#0,i(a5)
	move.w	d4,d2		; d2.w = v
	bpl.s	psx
	neg.w	d2
psx	move.l	d4,d1
	swap	d1		; d1.w = u
	tst.w	d1
	bpl.s	psy
	neg.w	d1
psy	cmp.w	d1,d2
	bgt.s	u_gr_v
	move.w	d1,m(a5)
	move.w	d2,n(a5)
	move.l	d0,d1
	clr.w	d1
	move.l	d1,d2x(a5)	; d2x and d2y
	bra.s	v_gr_u
u_gr_v
	move.w	d1,n(a5)
	move.w	d2,m(a5)
	clr.l	d1
	move.w	d0,d1
	move.l	d1,d2x(a5)	; d2x and d2y

v_gr_u	move.w	m(a5),d0
	lsr.w	#1,d0
	move.w	d0,s(a5)
	move.w	speeds(pc),d2	; Amount of different speeds
	bsr.s	random
	move.b	d0,speed(a5)	; Speed of this star

	move.b	scolor(pc),d1
	bne.s	fixcolor
	andi.l	#$ff,d0 	; The color depends on the speed of the star (slower->darker)
	asl.l	#2,d0
	divu	speeds(pc),d0
	moveq	#3,d1
	sub.b	d0,d1
fixcolor
	move.b	d1,color(a5)

	clr.b	speedcn(a5)

	move.w	#100,d2 	; Star is suppressed while this counter is > 0
	bsr.s	random		; Without this all stars would start in the screen center
	add.w	#10,d0
	move.b	d0,actnum(a5)

	move.w	#0,active(a5)
	rts

; d0.w : word to test / result
sgn	tst.w	d0
	beq.s	sgnexit
	bmi.s	sgnsign
	move.w	#1,d0
sgnexit rts
sgnsign move.w	#-1,d0
	rts

; d0.w : random number
; d2.w : range of random number
; a6.l : base of sysvars
random
	move.w	sys_rand(a6),d0 ; System random variable. It is updated by the
	mulu	#$a279,d0	; system 50 times per second. As that's too slow
	rol.w	#1,d0		; I took these 4 lines out of SMSQ/E (I don't
	move.w	d0,sys_rand(a6) ; know why the value is $A279)
	and.l	#$ffff,d0	; I only want a word to divide
	divu	d2,d0
	swap	d0		; Take the rest
	rts

	end
